// This Source Code Form is subject to the terms of the Mozilla Public
// License, v. 2.0. If a copy of the MPL was not distributed with this
// file, You can obtain one at http://mozilla.org/MPL/2.0/.

pub const OP_OK: u16 = 0;
pub const OP_ERR: u16 = 1;

pub const OP_OK_STR: &str = "OK";

pub const RPC_SOCKADDR: &str = "/run/stratisd/stratisd-min-jsonrpc";
